from ftw.contentstats.stats import ContentStats
from zope.publisher.browser import BrowserView
import json


class ContentStatsView(BrowserView):
    """Displays content statistics.
    """

    VISUALIZED_STATS = (
        'portal_types',
        'review_states',
        'checked_out_docs',
        'file_mimetypes',
    )

    def data_url_for(self, stat_name):
        """Build the data URL for a particular stat.
        """
        base_url = '/'.join(
            (self.context.absolute_url(), 'content-stats-json?stat=%s'))
        return base_url % stat_name

    def get_visualized_stats(self):
        """Used in template to render chart containers for each stat.

        Limits the returned stats to ones we want to visualize.

        The actual *data* from these stats is only used in template to render
        the HTML table for graceful degradation - the actual charts fetch
        their data from the JSON view below.
        """
        stats = {}
        all_stats = ContentStats().get_human_readable_stats()

        # Filter out stats not marked for visualization and
        # inject respective data URL for each stat
        for name, stats_data in all_stats.items():
            if name not in self.VISUALIZED_STATS:
                continue

            stats_data['data_url'] = self.data_url_for(name)
            stats[name] = stats_data

        return stats.items()


class ContentStatsJSONView(BrowserView):
    """Return stats from a particular stats provider as JSON.

    Used to fetch data via data.url from the C3 charts. This will return the
    human readable representation of the stats.
    """

    def __call__(self):
        self.request.response.setHeader('Content-Type', 'application/json')

        stat_name = self.request.form.get('stat')
        all_stats = ContentStats().get_human_readable_stats()

        try:
            stats = all_stats[stat_name]
        except KeyError:
            self.request.response.setStatus(404)
            return json.dumps({})

        return json.dumps(stats['data'])



from ftw.contentstats.tests import FunctionalTestCase
from plone import api


PROFILE = 'ftw.contentstats:default'


class TestDefaultProfile(FunctionalTestCase):

    def test_installed(self):
        portal_setup = api.portal.get_tool('portal_setup')
        version = portal_setup.getLastVersionForProfile(PROFILE)
        self.assertNotEqual(version, None)
        self.assertNotEqual(version, 'unknown')



from datetime import datetime
from ftw.contentstats.disk_usage import DiskUsageCalculator
from ftw.contentstats.tests import assets
from ftw.testing import freeze
from textwrap import dedent
from unittest import TestCase
import json


FROZEN_NOW = datetime(2018, 12, 30, 12, 45)


class TestDiskUsageCalculator(TestCase):

    def setUp(self):
        self.deployment_path = '/path/to/deployment'

    def test_parse_du_output_total(self):
        calculator = DiskUsageCalculator(self.deployment_path)
        output_total = dedent("""\
        1024   /path/to/deployment
        """)
        disk_usage = calculator.parse_du_output(output_total)
        self.assertEqual({'': 1024}, disk_usage)

    def test_parse_du_output_subtrees(self):
        calculator = DiskUsageCalculator(self.deployment_path)
        output_subtrees = dedent("""\
        5  /path/to/deployment/bin/instance
        10 /path/to/deployment/bin
        15 /path/to/deployment/var/log
        20 /path/to/deployment/var/filestorage/Data.fs
        25 /path/to/deployment/var/filestorage/Data.fs.index
        35 /path/to/deployment/var/blobstorage/0x00
        40 /path/to/deployment/var/blobstorage/tmp
        45 /path/to/deployment/var/blobstorage
        50 /path/to/deployment
        """)
        disk_usage = calculator.parse_du_output(output_subtrees)
        self.assertEqual({'': 50,
                          'bin': 10,
                          'bin/instance': 5,
                          'var/blobstorage': 45,
                          'var/blobstorage/0x00': 35,
                          'var/blobstorage/tmp': 40,
                          'var/filestorage/Data.fs': 20,
                          'var/filestorage/Data.fs.index': 25,
                          'var/log': 15},
                         disk_usage)

    def test_parse_du_output_strips_whitespace(self):
        calculator = DiskUsageCalculator(self.deployment_path)
        output_whitespace = """

        1024   /path/to/deployment/foo
        1024   \t\t/path/to/deployment/bar
        2048      /path/to/deployment\t

        """
        disk_usage = calculator.parse_du_output(output_whitespace)
        self.assertEqual({'': 2048,
                          'bar': 1024,
                          'foo': 1024},
                         disk_usage)

    def test_calc_du_total(self):
        calculator = DiskUsageCalculator(self.deployment_path)
        calculator.du_outputs['total'] = dedent("""\
        1024   /path/to/deployment
        """)
        disk_usage = calculator.calc_du_total()
        self.assertEqual(1024, disk_usage)

    def test_calc_du_subtrees(self):
        calculator = DiskUsageCalculator(self.deployment_path)
        calculator.du_outputs['subtrees'] = dedent("""\
        5  /path/to/deployment/bin/instance
        10 /path/to/deployment/bin
        15 /path/to/deployment/var/log
        20 /path/to/deployment/var/filestorage/Data.fs
        25 /path/to/deployment/var/filestorage/Data.fs.index
        35 /path/to/deployment/var/blobstorage/0x00
        40 /path/to/deployment/var/blobstorage/tmp
        45 /path/to/deployment/var/blobstorage
        50 /path/to/deployment
        """)
        disk_usage = calculator.calc_du_subtrees()

        # Should reduce subtree depth to 2 levels, except for var/filestorage/*
        self.assertEqual({'bin': 10,
                          'bin/instance': 5,
                          'var/blobstorage': 45,
                          'var/filestorage/Data.fs': 20,
                          'var/filestorage/Data.fs.index': 25,
                          'var/log': 15},
                         disk_usage)

    def test_calc_du_stats(self):
        calculator = DiskUsageCalculator(self.deployment_path)

        calculator.du_outputs['subtrees'] = dedent("""\
        5  /path/to/deployment/bin/instance
        10 /path/to/deployment/bin
        15 /path/to/deployment/var/log
        20 /path/to/deployment/var/filestorage/Data.fs
        25 /path/to/deployment/var/filestorage/Data.fs.index
        35 /path/to/deployment/var/blobstorage/0x00
        40 /path/to/deployment/var/blobstorage/tmp
        45 /path/to/deployment/var/blobstorage
        50 /path/to/deployment
        """)

        calculator.du_outputs['total'] = dedent("""\
        1024   /path/to/deployment
        """)

        with freeze(FROZEN_NOW):
            du_stats = calculator.calc_du_stats()

        # Should reduce subtree depth to 2 levels, except for var/filestorage/*
        self.assertEqual({'blobstorage': 45,
                          'deployment': '/path/to/deployment',
                          'filestorage': 20,
                          'updated': FROZEN_NOW.isoformat(),
                          'total': 1024,
                          'subtrees': {'bin': 10,
                                       'bin/instance': 5,
                                       'var/blobstorage': 45,
                                       'var/filestorage/Data.fs': 20,
                                       'var/filestorage/Data.fs.index': 25,
                                       'var/log': 15},
                          },
                         du_stats)

        # Ensure the du_stats format matches our asset file which is also
        # used to test the provider. Since this is the interchange format
        # between the bin/dump-content-stats script that dumps it and the
        # @@dump-content-stats view that reads it, we need to make sure
        # these implementations stay in sync.
        asset = json.loads(assets.load('disk-usage.json'))
        self.assertEqual(asset, du_stats)



from ftw.contentstats.interfaces import IStatsProvider
from ftw.contentstats.logger import get_log_dir_path
from logging import getLogger
from os.path import join
from Products.CMFPlone.interfaces import IPloneSiteRoot
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface
import json


log = getLogger()


@implementer(IStatsProvider)
@adapter(IPloneSiteRoot, Interface)
class DiskUsageProvider(object):
    """Reads disk usage statistics from a file in var/log/disk-usage.json.

    This file gets created/updated by the bin/dump-content-stats script.
    """

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def title(self):
        """Human readable title
        """
        return u'Disk usage statistics'

    def get_raw_stats(self):
        """Returns a dict with disk usage stats.
        """
        log_dir = get_log_dir_path()
        path = join(log_dir, 'disk-usage.json')
        try:
            with open(path) as disk_usage_file:
                disk_usage = json.load(disk_usage_file)
        except IOError:
            log.warn('Unable to read disk usage stats from %r' % path)
            return {}

        stats = {}
        stats['total'] = disk_usage['total']
        stats['filestorage'] = disk_usage['filestorage']
        stats['blobstorage'] = disk_usage['blobstorage']
        return stats

    def get_display_names(self):
        return None



from ftw.contentstats.disk_usage import DiskUsageCalculator
from ftw.contentstats.logger import log_stats_to_file
from zope.component.hooks import setSite
import App.config
import argparse
import os.path
import sys


def get_deployment_path():
    cfg = App.config._config
    return os.path.normpath(os.path.join(cfg.clienthome, '..', '..'))


def dump_content_stats(app, args):
    parser = argparse.ArgumentParser(description="Dump content stats.")
    parser.add_argument(
        '--site-path', '-s',
        help='Path to the Plone site.',
        default=None,
    )
    parser.add_argument(
        '--python-du',
        help='Use Python implementation for disk usage calculation.',
        action='store_true',
    )
    parser.add_argument(
        '--data-path', '-d',
        help='Path to data directory for which to calculate disk usage.',
        default=None,
    )
    parser.add_argument(
        '--filestorage-path',
        help='Path to Data.fs used for filestorage disk usage.',
        default='var/filestorage/Data.fs',
    )
    parser.add_argument(
        '--blobstorage-path',
        help='Path to directory used for blobstorage disk usage.',
        default='var/blobstorage',
    )

    if sys.argv[0] != 'dump_content_stats':
        args = args[2:]
    options = parser.parse_args(args)

    deployment_path = get_deployment_path()
    DiskUsageCalculator(
        deployment_path,
        use_du_util=not options.python_du,
        data_path=options.data_path,
        filestorage_path=options.filestorage_path,
        blobstorage_path=options.blobstorage_path,
    ).calc_and_dump()

    site = setup_site(app, options)
    if site is None:
        sys.exit(1)

    log_stats_to_file()


def get_site(app, options):
    if options.site_path:
        return app.unrestrictedTraverse(options.site_path)
    else:
        sites = []
        for item in app.values():
            if item.meta_type == 'Plone Site':
                sites.append(item)

        if len(sites) == 1:
            return sites[0]
        elif len(sites) > 1:
            print('ERROR: Multiple Plone sites found. Use -s to specify site.')
        else:
            print('ERROR: No Plone site found.')


def setup_site(app, options):
    # Delay import of the Testing module
    # Importing it before the database is opened, will result in opening a
    # DemoStorage database instead of the one from the config file.
    from Testing.makerequest import makerequest  # noqa
    app = makerequest(app)
    site = get_site(app, options)
    if site is not None:
        setSite(site)
    return site



from distutils.spawn import find_executable
from ftw.contentstats.disk_usage import DiskUsageCalculator
from ftw.contentstats.utils import get_buildout_path
from ftw.contentstats.utils import get_zope_url
import argparse
import os
import requests
import sys


def parse_args():
    parser = argparse.ArgumentParser(
        description="""
            Dump content stats to logfile.

            Will dump content stats to logfile by making a request to
            the @@dump-content-stats view.

            Invoke this command on the same machine that a Plone site is
            running in order to have content stats dumped to the JSON logfile.
            """)
    parser.add_argument(
        '--site-id', '-s',
        help='Path to the Plone site.',
        required=True)
    parser.add_argument(
        '--python-du',
        help='Use Python implementation for disk usage calculation.',
        action='store_true',
    )
    parser.add_argument(
        '--data-path', '-d',
        help='Path to data directory for which to calculate disk usage.',
        default=None,
    )
    parser.add_argument(
        '--filestorage-path',
        help='Path to Data.fs used for filestorage disk usage.',
        default='var/filestorage/Data.fs',
    )
    parser.add_argument(
        '--blobstorage-path',
        help='Path to directory used for blobstorage disk usage.',
        default='var/blobstorage',
    )

    args = parser.parse_args()
    return args


def renice():
    """Set CPU and I/O scheduling priorities so that this process
    (the bin/dump-content-stats script, not the Plone instances) run with
    lower priority and give way to other processes if required.
    """
    os.nice(10)
    ionice_path = find_executable('ionice')

    if ionice_path is None:
        print "Unable to find 'ionice' executable, skipping ionicing..."
        return

    # class 2: Best-effort, prio 7: lowest (in that class)
    ionice_cmd = '%s -c2 -n7 -p %s' % (ionice_path, os.getpid())
    os.system(ionice_cmd)


def dump_stats_cmd():
    """Will dump content stats to logfile via the @@dump-content-stats view.
    """
    args = parse_args()

    # Lower CPU and I/O scheduling priority
    renice()

    # Calculate disk usage stats and dump them to var/log/disk-usage.json
    deployment_path = get_buildout_path()
    DiskUsageCalculator(
        deployment_path,
        use_du_util=not args.python_du,
        data_path=args.data_path,
        filestorage_path=args.filestorage_path,
        blobstorage_path=args.blobstorage_path,
    ).calc_and_dump()

    zope_url = get_zope_url()
    plone_url = ''.join((zope_url, args.site_id))
    dump_stats_url = '/'.join((plone_url, '@@dump-content-stats'))

    with requests.Session() as session:
        response = session.post(dump_stats_url)
        if response.status_code == 204:
            print "Stats dumped"
            sys.exit(0)

        else:
            print "Failed to dump stats:"
            print "Got response %r for URL %r" % (response, dump_stats_url)
            print
            print response.content
            print repr(response)
            sys.exit(1)



"""Contains functions to calculate disk usage stats of a deployment.

These are intended to be run from the bin/dump-content-stats script (outside
a Plone process), and write the resulting stats to var/log/disk-usage.json
in order to be picked up by the @@dump-content-stats view later.
"""

from datetime import datetime
from distutils.spawn import find_executable
from subprocess import check_output
import json
import os
import shlex
import sys


try:
    from scandir import walk
except ImportError:
    from os import walk


class DiskUsageCalculator(object):

    du_stats_path = 'var/log/disk-usage.json'

    def __init__(
        self,
        deployment_path,
        use_du_util=True,
        data_path=None,
        filestorage_path='var/filestorage/Data.fs',
        blobstorage_path='var/blobstorage',
    ):
        self.deployment_path = deployment_path
        self.du_executable = None
        self.use_du_util = use_du_util
        self.data_path = data_path or deployment_path
        self.filestorage_path = filestorage_path
        self.blobstorage_path = blobstorage_path

        # This facilitates testing without actually running `du`
        self.du_outputs = {}

    def calc_and_dump(self):
        """Calculate disk usage stats and dump them.

        Results will be written to a JSON file in var/log/ for pickup by the
        @@dump-content-stats view (we don't want to run I/O heavy and potentially
        long running jobs like this from inside a Zope instance).
        """
        if self.use_du_util:
            self.du_executable = get_du_executable()

        du_stats = self.calc_du_stats()
        self.dump(du_stats)

    def disk_usage(self, path='.', max_depth=3):
        """Get disk usage for the given path.

           Calculates totals for up to max_depth levels.
           Return tuples of size counting all hardlinks, size counting hardlinks
           only once, blocks counting all hardlinks and blocks counting hardlinks
           only once. 
        """
        seen_inodes = set()
        results = {}
        for dirpath, dirnames, filenames in walk(path):
            total_size = 0      # Size counting all hardlinks
            total_size_h = 0    # Size counting hardlinks only once
            total_blocks = 0    # Blocks counting all hardlinks
            total_blocks_h = 0  # Blocks counting hardlinks only once
            for filename in filenames:
                fp = os.path.join(dirpath, filename)
                stat_result = os.lstat(fp)
                size = stat_result.st_size
                size_h = 0
                blocks = stat_result.st_blocks
                blocks_h = 0
                if stat_result.st_nlink > 1:
                    if stat_result.st_ino not in seen_inodes:
                        seen_inodes.add(stat_result.st_ino)
                        size_h = size
                        blocks_h = blocks
                else:
                    size_h = size
                    blocks_h = blocks
                total_size += size
                total_size_h += size_h
                total_blocks += blocks
                total_blocks_h += blocks_h
                rel_path = os.path.relpath(fp, start=path)
                if len(rel_path.split(os.sep)) <= max_depth:
                    results[rel_path] = (size, size_h, blocks, blocks_h)
            rel_path = os.path.relpath(dirpath, start=path)
            if rel_path != '.':
                rel_path = os.path.join('.', rel_path)
            sub_path = ''
            for path_seg in rel_path.split(os.sep)[:max_depth + 1]:
                sub_path = os.path.normpath(os.path.join(sub_path, path_seg))
                s, sh, b, bh = results.setdefault(sub_path, (0, 0, 0, 0))
                results[sub_path] = (
                    s + total_size,
                    sh + total_size_h,
                    b + total_blocks,
                    bh + total_blocks_h,
                )
        return results

    def calc_du_stats(self):
        """Determine disk usage (using `du`) for Data.fs and blobstorage.

        Returns a dictionary with separate keys for total deployment size and
        individual subtrees.
        """
        if self.du_executable is None and not self.du_outputs:
            res = self.disk_usage(path=self.data_path)
            disk_usage_total = res['.'][1]
            disk_usage_subtrees = {k: v[0] for k, v in res.items()}
        else:
            # Calculate an accurate total first
            disk_usage_total = self.calc_du_total()

            # Calculate size of individual subdirectories / files
            disk_usage_subtrees = self.calc_du_subtrees()

        du_stats = {}
        du_stats['deployment'] = self.deployment_path
        du_stats['updated'] = datetime.now().isoformat()
        du_stats['total'] = disk_usage_total
        du_stats['subtrees'] = disk_usage_subtrees
        du_stats['filestorage'] = disk_usage_subtrees.get(self.filestorage_path, 0)
        du_stats['blobstorage'] = disk_usage_subtrees.get(self.blobstorage_path, 0)
        return du_stats

    def calc_du_total(self):
        """Calculate an accurate total for the deployment directory.

        Sum up the size of the entire deployment directory, counting hardlinked
        files only once, and using actual FS block usage (not apparent size).
        This correctly reports the real disk usage of the deployment, accounting
        for hardlinks, sparse files and files smaller than the FS block size.

        Returns the total size in bytes as an integer.
        """
        if 'total' not in self.du_outputs:
            du_total_cmd = '%s -s -B1 %s' % (self.du_executable, self.data_path)
            self.du_outputs['total'] = self.run(du_total_cmd)

        disk_usage_total = self.parse_du_output(self.du_outputs['total'])

        # We have exactly one result line, the deployment directory itself
        assert disk_usage_total.keys() == ['']
        return disk_usage_total['']

    def calc_du_subtrees(self):
        """Calculate size of individual subdirectories / files in deployment dir.

        This is needed to report database size (Data.fs, blobstorage) without
        accounting for hardlinks. We actually want to count hardlinks twice,
        because otherwise we would be dependent on `du` counting blobstorage
        and blobstorage.old in exactly the right order.

        This listing might eventually also be used by agent.smith to replace
        the `du/big` command.
        """
        if 'subtrees' not in self.du_outputs:
            du_subtrees_cmd = '%s -x -a -B1 --count-links --apparent-size --max-depth=3 %s' % (
                self.du_executable, self.data_path)
            self.du_outputs['subtrees'] = self.run(du_subtrees_cmd)

        disk_usage_subtrees = self.parse_du_output(self.du_outputs['subtrees'])

        # Remove total, because it's inaccurate in this run
        disk_usage_subtrees.pop('')

        # Reduce subtree depth to 2, except for var/filestorage/*
        # (We need the size of Data.fs, which is 3 levels deep)
        for key in disk_usage_subtrees.keys():
            path = key.lstrip('/')
            if path.count(os.sep) > 1 and 'filestorage' not in path:
                disk_usage_subtrees.pop(key)
        return disk_usage_subtrees

    def parse_du_output(self, output):
        """Parse output of a 'du ...' command.

        Returns a dict (path -> size_in_bytes) with paths relative to the deployment
        directory.
        """
        disk_usage = {}
        lines = map(str.strip, output.strip().splitlines())
        for line in lines:
            size_in_bytes, path = map(str.strip, line.split())
            size_in_bytes = int(size_in_bytes)
            assert path.startswith(self.data_path)
            path = path.replace(self.data_path, '', 1).lstrip('/')
            disk_usage[path] = size_in_bytes

        return disk_usage

    def run(self, cmd):
        return check_output(shlex.split(cmd))

    def dump(self, du_stats):
        """Dump disk usage to file for pickup by @@dump-content-stats view
        (or, eventually, agent.smith)
        """
        du_stats_path = os.path.join(self.deployment_path, self.du_stats_path)
        with open(du_stats_path, 'w') as du_stats_file:
            json.dump(du_stats, du_stats_file)


def get_du_executable():
    """Get path to `du` (Linux) or `gdu` (Mac OS).

    The flags we need for `du` are not supported by the default *BSD-style
    `du` that is shipped with Mac OS. However, the GNU-style `du` can be
    installed on Mac OS (as `gdu`), so we attempt to look for it's presence.
    This way, the disk usage aspect can still be tested locally on Mac OS.
    """
    if sys.platform == 'darwin':
        du_executable = find_executable('gdu')
        if du_executable is None:
            print "WARNING: Unable to locate `gdu` executable."
            print
            print "Disk usage stats will be computed using internal implementation."
            print
            print "If you want to use the 'du' utility on Mac OS, you need"
            print "to install GNU coreutils to get the GNU-style du command"
            print "(installed as `gdu`):"
            print
            print "  brew install coreutils"
            print

        return du_executable

    else:
        du_executable = find_executable('du')
        return du_executable



from ftw.builder import Builder
from ftw.builder import create
from ftw.contentstats.stats import ContentStats
from ftw.contentstats.tests import FunctionalTestCase
from ftw.testbrowser import browsing
from zExceptions import Unauthorized


class TestContentStatsView(FunctionalTestCase):

    def setUp(self):
        super(TestContentStatsView, self).setUp()
        self.grant('Manager')

    def create_content(self):
        self.set_workflow_chain('Document', 'simple_publication_workflow')
        self.set_workflow_chain('Folder', 'simple_publication_workflow')
        create(Builder('folder'))
        create(Builder('page'))
        create(Builder('page')
               .in_state('published'))

    def test_content_stats_view_only_accessible_for_manager(self):
        self.grant('Contributor', 'Editor', 'Reviewer', 'Publisher')
        with self.assertRaises(Unauthorized):
            self.portal.restrictedTraverse('@@content-stats')

    @browsing
    def test_view_lists_type_counts_in_table(self, browser):
        self.create_content()
        browser.login().open(self.portal, view='@@content-stats')
        table = browser.css('#content-stats-portal_types').first
        self.assertEqual(
            [['', 'Folder', '1'], ['', 'Page', '2']],
            table.lists())

    @browsing
    def test_view_lists_review_states_in_table(self, browser):
        self.create_content()
        browser.login().open(self.portal, view='@@content-stats')
        table = browser.css('#content-stats-review_states').first
        self.assertEqual(
            [['', 'private', '2'], ['', 'published', '1']],
            table.lists())

    @browsing
    def test_editable_border_disabled(self, browser):
        browser.login().open(self.portal, view='@@content-stats')
        self.assertEqual(0, len(browser.css('#content-views')))

    @browsing
    def test_data_attribute_with_data_url(self, browser):
        browser.login().open(self.portal, view='@@content-stats')
        self.assertEqual(
            'http://nohost/plone/content-stats-json?stat=portal_types',
            browser.css('#content-stats-data-portal_types').first.attrib[
                'data-stat-data-url'])

    @browsing
    def test_json_endpoint_returns_404_for_missing_stat(self, browser):
        self.create_content()
        browser.login()

        with browser.expect_http_error(code=404):
            browser.open(self.portal, view='content-stats-json')

        self.assertEquals('application/json',
                          browser.headers.get('Content-Type'))

        self.assertDictEqual({}, browser.json)

    @browsing
    def test_json_endpoint_stats_by_name(self, browser):
        self.create_content()
        browser.login()

        browser.open(self.portal, view='content-stats-json?stat=portal_types')

        self.assertEquals('application/json',
                          browser.headers.get('Content-Type'))

        self.assertDictEqual(
            ContentStats().get_human_readable_stats()['portal_types']['data'],
            browser.json)



from ftw.builder import Builder
from ftw.builder import create
from ftw.contentstats.testing import CONTENTSTATS_FUNCTIONAL_ZSERVER
from ftw.contentstats.testing import FTW_MONITOR_INSTALLED
from ftw.contentstats.tests import FunctionalTestCase
from ftw.testbrowser import browsing
from requests_toolbelt.adapters.source import SourceAddressAdapter
import sys
import unittest


class TestContentStatsView(FunctionalTestCase):

    layer = CONTENTSTATS_FUNCTIONAL_ZSERVER

    def setUp(self):
        super(TestContentStatsView, self).setUp()
        self.grant('Anonymous')

    def create_content(self):
        self.set_workflow_chain('Document', 'simple_publication_workflow')
        self.set_workflow_chain('Folder', 'simple_publication_workflow')
        create(Builder('folder'))
        create(Builder('page'))
        create(Builder('page')
               .in_state('published'))

    @browsing
    def test_causes_stats_to_be_logged(self, browser):
        self.grant('Contributor')
        self.create_content()
        self.grant('Anonymous')

        browser.open(self.portal, view='@@dump-content-stats', method='POST')
        log_entry = self.get_log_entries()[-1]

        expected_stats = [
            'site',
            'disk_usage',
            'portal_types',
            'review_states',
            'timestamp',
        ]
        if FTW_MONITOR_INSTALLED:
            expected_stats.append('perf_metrics')

        self.assertItemsEqual(expected_stats, log_entry.keys())

        self.assertEquals(
            {u'Folder': 1, u'Document': 2},
            log_entry['portal_types'])

        self.assertEquals({}, log_entry['disk_usage'])

    @browsing
    def test_access_from_localhost_allowed_for_anonymous(self, browser):
        browser.open(self.portal, view='@@dump-content-stats', method='POST')
        self.assertEquals(204, browser.status_code)

    # Mac OS rejects source addresses other than 127.0.0.1 from the loopback
    # interface with "[Errno 49] Can't assign requested address"
    @unittest.skipIf(sys.platform == 'darwin', "Can't test this on Mac OS")
    @browsing
    def test_only_allows_access_from_localhost(self, browser):
        source = SourceAddressAdapter('127.0.0.42')
        browser.get_driver().requests_session.mount('http://', source)
        dump_stats_url = (
            'http://localhost:%s/%s/@@dump-content-stats' % (
                self.zserver_port, self.portal.id))

        with browser.expect_unauthorized():
            browser.open(dump_stats_url)

        with browser.expect_unauthorized():
            browser.open(dump_stats_url, method='POST')

    @browsing
    def test_only_accepts_post_method(self, browser):
        with browser.expect_http_error(code=405):
            browser.open(self.portal, view='@@dump-content-stats')

        browser.open(self.portal, view='@@dump-content-stats', method='POST')
        self.assertEquals(204, browser.status_code)



from ftw.builder import Builder
from ftw.builder import create
from ftw.contentstats.interfaces import IStatsProvider
from ftw.contentstats.stats import ContentStats
from ftw.contentstats.testing import FTW_MONITOR_INSTALLED
from ftw.contentstats.tests import FunctionalTestCase
from unittest import TestCase
from zope.interface.verify import verifyClass


class TestContentStatsNoPlone(TestCase):

    def test_raise_exception_if_there_is_no_plone(self):
        with self.assertRaises(Exception):
            ContentStats()


class TestContentStats(FunctionalTestCase):

    def setUp(self):
        super(TestContentStats, self).setUp()
        self.grant('Manager')

        self.stats_util = ContentStats()

    def create_content(self):
        self.set_workflow_chain('Document', 'simple_publication_workflow')
        self.set_workflow_chain('Folder', 'simple_publication_workflow')
        create(Builder('folder'))
        create(Builder('page'))
        create(Builder('page')
               .in_state('published'))

    def test_all_registered_providers_respect_the_contract(self):
        for name_, provider in self.stats_util._get_providers():
            verifyClass(IStatsProvider, provider.__class__)

    def test_get_all_provider_names(self):
        expected_provider_names = [
            'portal_types',
            'review_states',
            'disk_usage',
        ]
        if FTW_MONITOR_INSTALLED:
            expected_provider_names.append('perf_metrics')

        self.assertItemsEqual(expected_provider_names,
                              self.stats_util.get_provider_names())

    def test_stats_contain_portal_types_stats(self):
        self.create_content()
        stats = self.stats_util.get_human_readable_stats()

        self.assertIn('portal_types', stats)
        self.assertDictEqual(
            {u'Folder': 1, u'Page': 2},
            stats['portal_types']['data'])

        self.assertEquals(u'Portal type statistics',
                          stats['portal_types']['title'])

    def test_stats_contain_review_state_stats(self):
        self.create_content()

        stats = self.stats_util.get_human_readable_stats()
        self.assertIn('review_states', stats)
        self.assertDictEqual(
            {'private': 2, 'published': 1},
            stats['review_states']['data'])

        self.assertEquals(u'Review state statistics',
                          stats['review_states']['title'])

    def test_stats_contain_disk_usage_stats(self):
        stats = self.stats_util.get_human_readable_stats()
        self.assertIn('disk_usage', stats)

        self.assertEquals(u'Disk usage statistics',
                          stats['disk_usage']['title'])



from App.config import getConfiguration
from ftw.contentstats.interfaces import IStatsProvider
from ftw.contentstats.logger import root_logger as logger
from glob import glob
from os.path import basename
from os.path import join as pjoin
from Products.CMFPlone.interfaces import IPloneSiteRoot
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface
from ZServer.medusa.http_server import http_server
import json
import re
import socket


INSTANCE_NAME_PATTERN = re.compile(r'instance(\d+)')


@implementer(IStatsProvider)
@adapter(IPloneSiteRoot, Interface)
class PerformanceMetricsProvider(object):
    """Gathers performance metrics from ftw.monitor (if installed).
    """

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def title(self):
        return u'Performance metrics'

    def get_raw_stats(self):
        """Return a dict with performance metrics.
        """
        stats = {}

        try:
            monitor_ports = get_monitor_ports(self.request)
        except Exception as exc:
            logger.warn('Failed to determine monitor ports: %r' % exc)
            return {}

        monitor_ports = get_monitor_ports(self.request)
        logger.info('Selected monitor ports for dumping perf_metrics: %r' % monitor_ports)
        for instance_name, monitor_port in monitor_ports.items():
            try:
                reply = netcat('127.0.0.1', monitor_port,
                               'perf_metrics\n').strip()
                instance_metrics = json.loads(reply)
                stats[instance_name] = instance_metrics
            except Exception as exc:
                logger.warn('Failed to gather perf_metrics from %s: %r' % (instance_name, exc))

        return stats

    def get_display_names(self):
        return None


def get_monitor_ports(request):
    """Get a list of open ftw.monitor ports for instances of this deployment.

    This will:
    - Determine which bin/instance* scripts belong to this deployment
    - Determine the instance numbers from them
    - Determine the port base via the ZServer port of the current instance
    - Build the map of monitor ports by instance by doing port arithmetic:
      monitor_port = port base + instance number + 80
      (According to 4tw port schema and ftw.monitor port strategy)
    """
    config = getConfiguration()
    instance_scripts = get_instance_scripts(config)
    instance_names = get_instance_names(instance_scripts)

    zserver_port = get_zserver_port(config)
    if zserver_port == 8080:
        # Local development
        return {'instance': 8160}

    port_base = get_port_base(zserver_port)

    monitor_ports_by_instance = {}
    for name in instance_names:
        if name in ('instance0', 'instance'):
            # Ignore maintenance or development instances
            continue

        number = get_instance_number(name)
        if number:
            monitor_port = port_base + number + 80
            monitor_ports_by_instance[name] = monitor_port

    return monitor_ports_by_instance


def get_buildout_dir(config):
    buildout_dir = config.instancehome.partition('/parts/')[0]
    return buildout_dir


def get_instance_scripts(config):
    buildout_dir = get_buildout_dir(config)
    bin_dir = pjoin(buildout_dir, 'bin')
    instance_scripts = glob(pjoin(bin_dir, 'instance*'))
    return instance_scripts


def get_instance_names(instance_scripts):
    return [basename(path) for path in instance_scripts]


def get_instance_number(instance_name):
    match = INSTANCE_NAME_PATTERN.match(instance_name)
    if match:
        number = int(match.group(1))
        return number


def get_zserver_port(config):
    zservers = [
        server for server in config.servers
        if isinstance(server, http_server)
    ]

    zserver_ports = [zs.port for zs in zservers]

    assert len(zserver_ports) == 1
    zserver_port = zserver_ports[0]
    return zserver_port


def get_port_base(zserver_port):
    return int(str(zserver_port)[:-2] + '00')


def netcat(hostname, port, content):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect((hostname, port))
    s.sendall(content)

    result = ''
    while 1:
        data = s.recv(1024)
        if data == '':
            break
        result += data
    s.close()
    return result



from ftw.contentstats.testing import CONTENTSTATS_FUNCTIONAL
from ftw.contentstats.testing import get_log_path
from mock import patch
from plone import api
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from unittest import TestCase
import json
import os
import transaction


class FunctionalTestCase(TestCase):
    layer = CONTENTSTATS_FUNCTIONAL

    def setUp(self):
        self.portal = self.layer['portal']
        self.request = self.layer['request']
        self.load_zcml_string = self.layer['load_zcml_string']

        # Prevent get_monitor_ports() from interfering with tests that
        # don't run on a ftw.monitor layer
        self.patched_get_monitor_ports = patch(
            'ftw.contentstats.providers.perf_metrics.get_monitor_ports'
        ).__enter__()

    def tearDown(self):
        self.patched_get_monitor_ports.__exit__()

    def grant(self, *roles):
        setRoles(self.portal, TEST_USER_ID, list(roles))
        transaction.commit()

    def set_workflow_chain(self, for_type, to_workflow):
        wftool = api.portal.get_tool('portal_workflow')
        wftool.setChainForPortalTypes((for_type,),
                                      (to_workflow,))

    def get_log_entries(self):
        log_path = get_log_path()
        with open(log_path) as log:
            entries = map(json.loads, log.readlines())
        return entries

    @property
    def zserver_port(self):
        return str(os.environ.get('ZSERVER_PORT', 55001))



from ftw.contentstats.interfaces import IStatsKeyFilter
from Products.CMFPlone.interfaces import IPloneSiteRoot
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface


@implementer(IStatsKeyFilter)
@adapter(IPloneSiteRoot, Interface)
class PortalTypesFilter(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def keep(self, key):
        whitelist = ['Document', 'Event']
        if key in whitelist:
            return True
        return False


@implementer(IStatsKeyFilter)
@adapter(IPloneSiteRoot, Interface)
class ReviewStatesFilter(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def keep(self, key):
        whitelist = ['published']
        if key in whitelist:
            return True
        return False



from pkg_resources import resource_filename
from pkg_resources import resource_string


def load(asset_filename):
    return resource_string('ftw.contentstats.tests.assets', asset_filename)


def path_to_asset(asset_filename):
    return resource_filename('ftw.contentstats.tests.assets', asset_filename)



from ftw.contentstats.interfaces import IStatsKeyFilter
from ftw.contentstats.interfaces import IStatsProvider
from plone import api
from Products.CMFPlone.interfaces import IPloneSiteRoot
from zope.component import adapter
from zope.component import queryMultiAdapter
from zope.i18n import translate
from zope.interface import implementer
from zope.interface import Interface


@implementer(IStatsProvider)
@adapter(IPloneSiteRoot, Interface)
class ReviewStatesProvider(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def title(self):
        """Human readable title
        """
        return u'Review state statistics'

    def _get_filter(self):
        return queryMultiAdapter((self.context, self.request),
                                 IStatsKeyFilter, name='review_states')

    def get_raw_stats(self):
        """Return a list of dicts (keys: name, amount).
        """
        counts = {}
        key_filter = self._get_filter()
        catalog = api.portal.get_tool('portal_catalog')
        index = catalog._catalog.indexes['review_state']
        for key in index.uniqueValues():
            if key_filter and not key_filter.keep(key):
                continue
            t = index._index.get(key)
            if not isinstance(t, int):
                counts[key] = len(t)
            else:
                counts[key] = 1
        return counts

    def get_display_names(self):
        """Return a id, title mapping of all workflow state titles to use
        as display names.
        """
        catalog = api.portal.get_tool('portal_catalog')
        index = catalog._catalog.indexes['review_state']

        titles = [
            (wfstate, translate(
                wfstate, domain='plone', context=self.request))
            for wfstate in index.uniqueValues()]
        return dict(titles)



from ftw.builder import Builder
from ftw.builder import create
from ftw.contentstats.interfaces import IStatsProvider
from ftw.contentstats.tests import FunctionalTestCase
from zope.component import getMultiAdapter


class TestPortalTypesProvider(FunctionalTestCase):

    def setUp(self):
        super(TestPortalTypesProvider, self).setUp()
        self.grant('Manager')

        self.provider = getMultiAdapter((self.portal, self.portal.REQUEST),
                                        IStatsProvider,
                                        name='portal_types')

    def create_content(self):
        create(Builder('folder'))
        create(Builder('page'))
        create(Builder('page'))

    def test_type_counts_empty(self):
        counts = self.provider.get_raw_stats()
        self.assertEqual({}, counts)

    def test_type_counts_reported_correctly(self):
        self.create_content()
        counts = self.provider.get_raw_stats()
        self.assertEqual({u'Folder': 1, u'Document': 2}, counts)

    def test_display_names_reported_correctly(self):
        titles = self.provider.get_display_names()
        self.assertDictContainsSubset({
            'Discussion Item': u'Comment',
            'Document': u'Page',
            'News Item': u'News Item'},
            titles)

    def test_respects_portal_types_filter(self):
        self.create_content()
        self.load_zcml_string('<include package="ftw.contentstats.demo" />')
        counts = self.provider.get_raw_stats()
        self.assertEqual({u'Document': 2}, counts)



from App.config import getConfiguration
from ftw.builder.testing import BUILDER_LAYER
from ftw.builder.testing import functional_session_factory
from ftw.builder.testing import set_builder_session_factory
from ftw.contentstats.logger import setup_logger
from ftw.testbrowser import REQUESTS_BROWSER_FIXTURE
from ftw.testing.layer import COMPONENT_REGISTRY_ISOLATION
from logging import FileHandler
from pkg_resources import DistributionNotFound
from pkg_resources import get_distribution
from plone.app.testing import applyProfile
from plone.app.testing import FunctionalTesting
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from plone.testing import z2
from StringIO import StringIO
from zope.configuration import xmlconfig
import logging
import os
import pytz
import shutil
import tempfile
import ZConfig


IS_PLONE_5_OR_GREATER = get_distribution('Plone').version >= '5'


try:
    get_distribution('ftw.monitor')
    FTW_MONITOR_INSTALLED = True
except DistributionNotFound:
    FTW_MONITOR_INSTALLED = False


def get_log_path():
    """Get filesystem path to ftw.contentstats' logfile.
    """
    logger = setup_logger()
    handler = logger.handlers[0]
    if isinstance(handler, FileHandler):
        log_path = handler.stream.name
        return log_path


def clear_log_handlers():
    logger = logging.getLogger('ftw.contentstats')
    map(logger.removeHandler, logger.handlers)


class PatchedLogTZ(object):
    """Context manager that patches LOG_TZ to a given timezone.
    """

    def __init__(self, tzname):
        self.new_tz = pytz.timezone(tzname)

    def __enter__(self):
        from ftw.contentstats import logger
        self._original_log_tz = logger.LOG_TZ
        logger.LOG_TZ = self.new_tz
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        from ftw.contentstats import logger
        logger.LOG_TZ = self._original_log_tz


class ContentStatsLayer(PloneSandboxLayer):
    defaultBases = (PLONE_FIXTURE, BUILDER_LAYER)

    def setUp(self):
        # Keep track of temporary files we create
        self.tempdir = None
        super(ContentStatsLayer, self).setUp()

    def tearDown(self):
        super(ContentStatsLayer, self).tearDown()

        # Clean up all temporary files we created
        shutil.rmtree(self.tempdir)

    def mktemp(self):
        """Create a temporary file to use as the path for the eventlog.
        We don't actually care about the contents of this file, we just need
        it to get a valid writable path to pass to the eventlog config, so
        ftw.contentstats can derive its own logfile path from it.

        We create this in a known tempdir so we can also clean up the
        contentstats.json file at layer teardown.
        """
        self.tempdir = tempfile.mkdtemp()
        f = open(os.path.join(self.tempdir, 'instance0.log'), 'w')
        f.close()
        return f.name

    def setup_eventlog(self):
        """Create an eventlog ZConfig configuration and patch it onto the
        global config, so it's present when ftw.contentstats attempts to read
        it to derive its own logfile path from the eventlog's logfile path.
        """
        schema = ZConfig.loadSchemaFile(StringIO("""
            <schema>
              <import package='ZConfig.components.logger'/>
              <section type='eventlog' name='*' attribute='eventlog'/>
            </schema>
        """))

        fn = self.mktemp()
        eventlog_conf, handler = ZConfig.loadConfigFile(schema, StringIO("""
            <eventlog>
              <logfile>
                path %s
                level debug
              </logfile>
            </eventlog>
        """ % fn))

        assert eventlog_conf.eventlog is not None
        getConfiguration().eventlog = eventlog_conf.eventlog

    def remove_eventlog(self):
        conf = getConfiguration()
        del conf.eventlog

    def setUpZope(self, app, configurationContext):
        # Set up the eventlog config before setup_logger is called
        self.setup_eventlog()

        xmlconfig.string(
            '<configure xmlns="http://namespaces.zope.org/zope">'
            '  <include package="z3c.autoinclude" file="meta.zcml" />'
            '  <includePlugins package="plone" />'
            '  <includePluginsOverrides package="plone" />'
            '</configure>',
            context=configurationContext)

        z2.installProduct(app, 'ftw.contentstats')

        if not IS_PLONE_5_OR_GREATER:
            # The tests will fail with a
            # `ValueError: Index of type DateRecurringIndex not found` unless
            # the product 'Products.DateRecurringIndex' is installed.
            z2.installProduct(app, 'Products.DateRecurringIndex')

    def tearDownZope(self, app):
        self.remove_eventlog()

    def testTearDown(self):
        # Isolation: truncate ftw.contentstats' logfile after each test
        log_path = get_log_path()
        if log_path:
            with open(log_path, 'w') as f:
                f.truncate()

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'ftw.contentstats:default')
        if IS_PLONE_5_OR_GREATER:
            applyProfile(portal, 'plone.app.contenttypes:default')


CONTENTSTATS_FIXTURE = ContentStatsLayer()

CONTENTSTATS_FUNCTIONAL = FunctionalTesting(
    bases=(CONTENTSTATS_FIXTURE,
           COMPONENT_REGISTRY_ISOLATION,
           set_builder_session_factory(functional_session_factory)),
    name="ftw.contentstats:functional")

CONTENTSTATS_FUNCTIONAL_ZSERVER = FunctionalTesting(
    bases=(z2.ZSERVER_FIXTURE,
           REQUESTS_BROWSER_FIXTURE,
           CONTENTSTATS_FIXTURE,
           COMPONENT_REGISTRY_ISOLATION,
           set_builder_session_factory(functional_session_factory)),
    name="ftw.contentstats:functional:zserver")



from ftw.contentstats.interfaces import IStatsProvider
from ftw.contentstats.providers.perf_metrics import get_buildout_dir
from ftw.contentstats.providers.perf_metrics import get_instance_names
from ftw.contentstats.providers.perf_metrics import get_instance_number
from ftw.contentstats.providers.perf_metrics import get_port_base
from ftw.contentstats.testing import FTW_MONITOR_INSTALLED
from mock import Mock
from mock import patch
from plone.testing import Layer
from unittest import skipUnless
from unittest import TestCase
from ZODB.interfaces import IDatabase
from zope.component import getMultiAdapter
from zope.component import provideUtility


if FTW_MONITOR_INSTALLED:
    from ftw.monitor.testing import MONITOR_ZSERVER_TESTING
    from ftw.monitor.testing import MonitorTestCase
else:
    MonitorTestCase = TestCase
    MONITOR_ZSERVER_TESTING = Layer(name='stub')


@skipUnless(FTW_MONITOR_INSTALLED, 'Requires ftw.monitor to run')
class TestPerfMetricsProvider(MonitorTestCase):

    layer = MONITOR_ZSERVER_TESTING

    @patch('ftw.contentstats.providers.perf_metrics.get_monitor_ports')
    def test_reports_loads_and_stores(self, mocked_get_monitor_ports):
        mocked_get_monitor_ports.return_value = {
            'instance99': MonitorTestCase.MONITOR_PORT}

        # Plone's testing DB is named 'unnamed' in tests. Also register it
        # as 'main' in order for ftw.monitor to be able to pick it up
        db = self.layer['portal']._p_jar.db()
        provideUtility(db, IDatabase, name='main')

        provider = getMultiAdapter(
            (self.layer['portal'], self.portal.REQUEST),
            IStatsProvider, name='perf_metrics')

        perf_metrics = provider.get_raw_stats()

        actual_categories = perf_metrics['instance99'].keys()
        for expected_category in [u'instance', u'cache', u'db', u'memory']:
            self.assertIn(expected_category, actual_categories)


class TestPerfMetricsHelperFunctions(TestCase):

    def test_get_port_base(self):
        self.assertEqual(14700, get_port_base(14702))

    def test_get_instance_number(self):
        self.assertIsNone(get_instance_number('instance'))
        self.assertEqual(1, get_instance_number('instance1'))
        self.assertEqual(1, get_instance_number('instance01'))
        self.assertEqual(42, get_instance_number('instance42'))

    def test_get_instance_names(self):
        instance_scripts = ['bin/instance1', '/path/to/bin/instance2']
        self.assertEqual(['instance1', 'instance2'],
                         get_instance_names(instance_scripts))

    def test_get_buildout_dir(self):
        zconfig = Mock(instancehome='/apps/01-plone/parts/instance1')
        self.assertEqual('/apps/01-plone', get_buildout_dir(zconfig))



from contextlib import contextmanager
from datetime import timedelta
from fluent.asynchandler import FluentHandler
from freezegun import freeze_time
from ftw.builder import Builder
from ftw.builder import create
from ftw.contentstats.logger import get_log_dir_path
from ftw.contentstats.logger import log_stats_to_file
from ftw.contentstats.logger import setup_logger
from ftw.contentstats.testing import clear_log_handlers
from ftw.contentstats.testing import FTW_MONITOR_INSTALLED
from ftw.contentstats.testing import PatchedLogTZ
from ftw.contentstats.tests import assets
from ftw.contentstats.tests import FunctionalTestCase
from logging import FileHandler
from mock import patch
from operator import itemgetter
from os.path import join
import json
import os


@contextmanager
def env_var(name, value):
    assert name not in os.environ, \
        'Unexpectedly found the variable {} in the environment.'.format(name)
    os.environ[name] = value
    try:
        yield
    finally:
        os.environ.pop(name)


class TestLogging(FunctionalTestCase):

    def setUp(self):
        super(TestLogging, self).setUp()
        clear_log_handlers()
        log_dir = get_log_dir_path()
        self.disk_usage_path = join(log_dir, 'disk-usage.json')

    def tearDown(self):
        super(TestLogging, self).tearDown()
        try:
            os.unlink(self.disk_usage_path)
        except OSError:
            pass

    def create_content(self):
        self.set_workflow_chain('Document', 'simple_publication_workflow')
        self.set_workflow_chain('Folder', 'simple_publication_workflow')
        create(Builder('folder'))
        create(Builder('page'))
        create(Builder('page')
               .in_state('published'))

    def test_logs_raw_stats(self):
        self.grant('Contributor')
        self.create_content()
        self.grant('Anonymous')

        with open(self.disk_usage_path, 'w') as disk_usage_file:
            disk_usage_file.write(assets.load('disk-usage.json'))

        log_stats_to_file()
        log_entry = self.get_log_entries()[-1]

        expected_stats_names = [
            'site',
            'timestamp',
            'disk_usage',
            'portal_types',
            'review_states',
        ]

        if FTW_MONITOR_INSTALLED:
            expected_stats_names.append('perf_metrics')

        self.assertItemsEqual(expected_stats_names, log_entry.keys())

        self.assertEquals(
            {u'Folder': 1, u'Document': 2},
            log_entry['portal_types'])

        self.assertEquals(
            {u'private': 2, u'published': 1},
            log_entry['review_states'])

        self.assertEquals(
            {u'blobstorage': 45,
             u'filestorage': 20,
             u'total': 1024},
            log_entry['disk_usage'])

    def test_log_multiple_entries(self):
        # Frozen time is specified in UTC
        # tz_offset specifies what offset to UTC the local tz is supposed to
        # have. This is relevant for stdlib functions that return local times,
        # but *not* for ftw.contentstats, since we never fetch local times
        with freeze_time("2017-07-29 10:30:58.000750", tz_offset=7) as clock:
            with PatchedLogTZ('Europe/Zurich'):
                log_stats_to_file()
                clock.tick(timedelta(days=1))
                log_stats_to_file()

        log_entries = self.get_log_entries()

        self.assertEquals(2, len(log_entries))
        self.assertEquals(
            [u'2017-07-29T12:30:58.000750+02:00',
             u'2017-07-30T12:30:58.000750+02:00'],
            map(itemgetter('timestamp'), log_entries))

    def test_logs_plone_site_id(self):
        log_stats_to_file()
        log_entry = self.get_log_entries()[0]

        self.assertEquals(u'plone', log_entry['site'])

    def test_dst_rollover(self):
        # Start in winter (no DST), half an hour before switch to DST, which
        # will happen at 2017-03-26 01:00:00 UTC / 2017-03-26 02:00:00 CET
        # for Europe/Zurich
        with freeze_time("2017-03-26 00:30:00.000750", tz_offset=7) as clock, \
                PatchedLogTZ('Europe/Zurich'):
            log_stats_to_file()

            # No DST (winter) - UTC offset for Europe/Zurich should be +01:00
            log_entry = self.get_log_entries()[-1]
            self.assertEqual(u'2017-03-26T01:30:00.000750+01:00',
                             log_entry['timestamp'])

            # Forward one hour - rollover from winter to summer, it's now DST
            clock.tick(timedelta(hours=1))
            log_stats_to_file()

            # DST (summer) - UTC offset for Europe/Zurich should be +02:00,
            # and we "magically" skipped the hourd from 02:00 - 03:00
            log_entry = self.get_log_entries()[-1]
            self.assertEqual(u'2017-03-26T03:30:00.000750+02:00',
                             log_entry['timestamp'])

            # Fast forward to October, half an hour before end of DST
            clock.move_to("2017-10-29 00:30:00.000750")
            log_stats_to_file()

            # We're still just in DST (summer) - UTC offset for
            # Europe/Zurich should be +02:00
            log_entry = self.get_log_entries()[-1]
            self.assertEqual(u'2017-10-29T02:30:00.000750+02:00',
                             log_entry['timestamp'])

            # Forward one hour - rollover from summer to winter, DST ends
            clock.tick(timedelta(hours=1))
            log_stats_to_file()

            # No DST (winter) - UTC offset for Europe/Zurich should be +01:00,
            # and it's now "magically" 02:30 again, even though an hour passed
            log_entry = self.get_log_entries()[-1]
            self.assertEqual(u'2017-10-29T02:30:00.000750+01:00',
                             log_entry['timestamp'])


class TestFluentLogging(FunctionalTestCase):

    @patch('fluent.handler.FluentHandler.emit')
    def test_dispatches_to_fluentd(self, mock_emit):
        clear_log_handlers()

        with env_var('FLUENT_HOST', 'localhost'):
            log_stats_to_file()

        self.assertEqual(1, mock_emit.call_count)
        log_record = mock_emit.call_args[0][0]

        expected_stats_names = [
            'site',
            'timestamp',
            'disk_usage',
            'portal_types',
            'review_states',
        ]

        if FTW_MONITOR_INSTALLED:
            expected_stats_names.append('perf_metrics')

        self.assertItemsEqual(
            expected_stats_names,
            json.loads(log_record.msg).keys())

    @patch('ftw.contentstats.logger.get_logfile_path')
    def test_sets_up_file_handler_by_default(self, mocked_logpath):
        mocked_logpath.return_value = '/tmp/logfile'
        clear_log_handlers()

        logger = setup_logger()

        self.assertEqual(1, len(logger.handlers))
        handler = logger.handlers[0]
        self.assertIsInstance(handler, FileHandler)
        self.assertEqual('/tmp/logfile', handler.stream.name)

    @patch('ftw.contentstats.logger.get_logfile_path')
    def test_sets_up_fluent_handler_if_envvar_set(self, mocked_logpath):
        # Mock the presence of a possible log file path in order to test that
        # even if one could be determined, setup_logger() *doesn't* set up
        # a FileHandler if FLUENT_HOST is set.
        mocked_logpath.return_value = '/tmp/logfile'

        clear_log_handlers()

        with env_var('FLUENT_HOST', 'localhost'):
            logger = setup_logger()

        self.assertEqual(1, len(logger.handlers))
        handler = logger.handlers[0]
        self.assertIsInstance(handler, FluentHandler)

    def test_sets_up_tag_including_ns_for_fluent_handler(self):
        clear_log_handlers()

        with env_var('FLUENT_HOST', 'localhost'):
            with env_var('KUBERNETES_NAMESPACE', 'demo-example-org'):
                logger = setup_logger()

        self.assertEqual(1, len(logger.handlers))
        handler = logger.handlers[0]
        self.assertEqual('demo-example-org-contentstats-json.log', handler.tag)

    def test_sets_up_fallback_tag_for_fluent_handler(self):
        clear_log_handlers()

        with env_var('FLUENT_HOST', 'localhost'):
            logger = setup_logger()

        self.assertEqual(1, len(logger.handlers))
        handler = logger.handlers[0]
        self.assertEqual('contentstats-json.log', handler.tag)



from ftw.contentstats.interfaces import IStatsProvider
from zope.component import getAdapters
from zope.component.hooks import getSite


class ContentStats(object):
    """Gather content statistics from the Plone site.
    """

    def __init__(self):
        """This utility class requires a Plone site.
        """
        if not self.plone:
            raise Exception('Please setup a Plone site')

    @property
    def plone(self):
        """Get Plone site from globals
        """
        return getSite()

    def _get_providers(self):

        return getAdapters((self.plone, self.plone.REQUEST),
                           IStatsProvider)

    def get_provider_names(self):
        """Get names of all registered stats providers.
        """
        return [name for name, adapter_ in self._get_providers()]

    def get_raw_stats(self):
        """Get a dictionary with raw stats from all registered providers.

        This is the main API for accessing the machine readable representation
        of the collected stats.
        """
        stats = {}
        for name, provider in self._get_providers():
            stats[name] = provider.get_raw_stats()
        return stats

    def get_stats_titles(self):
        """Get a name:title mapping for titles of all providers.
        """
        titles = {}
        for name, provider in self._get_providers():
            titles[name] = provider.title()
        return titles

    def get_stats_display_names(self):
        """Get a name:display_names mapping with display_names dicts of
        all providers.

        If a provider returns None for its display_names mapping, this
        method substitutes it with an empty dict for easy processing below.
        """
        display_names = {}
        for name, provider in self._get_providers():
            display_names[name] = {}
            names = provider.get_display_names()
            if names:
                display_names[name] = names
        return display_names

    def get_human_readable_stats(self):
        """Get a dictionary that combines all stats with their metadata.

        This includes stat titles as well as rewriting internal keys to
        display names. This is used in the template and JSON view, where
        a human readable representation is desired.
        """
        raw_stats = self.get_raw_stats()
        titles = self.get_stats_titles()
        display_names = self.get_stats_display_names()

        human_readable_stats = {}
        for stat_name in self.get_provider_names():
            stat_dict = {}
            stat_dict['title'] = titles[stat_name]
            stat_dict['data'] = {}

            # Rewrite internal keys to display names if necessary
            names = display_names.get(stat_name, {})
            for key, value in raw_stats[stat_name].items():
                # Try for a display name, default to key
                display_name = names.get(key, key)
                stat_dict['data'][display_name] = value

            human_readable_stats[stat_name] = stat_dict

        return human_readable_stats



from contextlib import closing
from os.path import abspath
from os.path import join
from path import Path
import re
import socket
import sys


# XXX: These helper functions have been copied over from ftw.upgrade
#
# Unfortunately there's no trivial way to get the port of a potentially
# running Zope instance without resorting to bin/instance [zopectl_command]
# style commands.
#
# And those pose an operational risk when invoked by a scheduled job, because
# they'll indefinitely keep trying to connect to ZEO instead of terminating
# when ZEO is not running. This leads to more and more processes accumulating,
# and the system eventually running out of file descriptors, memory or other
# resources.


class NoRunningInstanceFound(Exception):
    pass


def get_buildout_path():
    # Path to bin/dump-content-stats script
    script_path = sys.argv[0]
    return Path(abspath(join(script_path, '..', '..')))


def get_zope_url():
    instance = get_running_instance(get_buildout_path())
    if not instance:
        raise NoRunningInstanceFound()
    return 'http://localhost:{0}/'.format(instance['port'])


def get_running_instance(buildout_path):
    for zconf in find_instance_zconfs(buildout_path):
        port = get_instance_port(zconf)
        if not port:
            continue
        if is_port_open(port):
            return {'port': port,
                    'path': zconf.dirname().dirname()}
    return None


def find_instance_zconfs(buildout_path):
    return sorted(buildout_path.glob('parts/*/etc/zope.conf'))


def get_instance_port(zconf):
    match = re.search(r'address ([\d.]*:)?(\d+)', zconf.text())
    if match:
        return int(match.group(2))
    return None


def is_port_open(port):
    result = -1
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        result = sock.connect_ex(('127.0.0.1', port))
    return result == 0



from setuptools import find_packages
from setuptools import setup
import os

version = '1.4.1'

tests_require = [
    'ftw.builder',
    'ftw.testbrowser',
    'ftw.testing',
    'mock',
    'plone.app.testing',
    'plone.testing',
    'plone.app.contenttypes',
    'transaction',
    'zope.configuration',
    'freezegun < 0.3.15',
    'requests_toolbelt',
]

extras_require = {
    'tests': tests_require,
    'scandir': ['scandir']
}


setup(
    name='ftw.contentstats',
    version=version,
    description='Collect and display content statistics for Plone sites',
    long_description=open("README.rst").read() + "\n" + open(
        os.path.join("docs", "HISTORY.txt")).read(),

    # Get more strings from
    # http://www.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Framework :: Plone',
        'Framework :: Plone :: 4.3',
        'Framework :: Plone :: 5.1',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],

    keywords='ftw content stats statistics',
    author='4teamwork AG',
    author_email='mailto:info@4teamwork.ch',
    url='https://github.com/4teamwork/ftw.contentstats',
    license='GPL2',

    packages=find_packages(exclude=['ez_setup']),
    namespace_packages=['ftw'],
    include_package_data=True,
    zip_safe=False,

    install_requires=[
        'Plone',
        'Products.CMFCore',
        'Products.CMFPlone',
        'Products.GenericSetup',
        'plone.api',
        'setuptools',
        'zope.i18nmessageid',
        'tzlocal',
        'pytz',
        'requests',
        'path.py >= 6.2',
        'fluent-logger',
    ],

    tests_require=tests_require,
    extras_require=extras_require,

    entry_points="""
    # -*- Entry points: -*-
    [z3c.autoinclude.plugin]
    target = plone

    [zopectl.command]
    dump_content_stats = ftw.contentstats.command:dump_content_stats

    [console_scripts]
    dump-content-stats = ftw.contentstats.console:dump_stats_cmd
    """,
)



from App.config import getConfiguration
from datetime import datetime
from fluent.asynchandler import FluentHandler
from fluent.handler import FluentRecordFormatter
from ftw.contentstats.stats import ContentStats
from logging import FileHandler
from logging import getLogger
from os.path import dirname
from os.path import join
from tzlocal import get_localzone
from zope.component.hooks import getSite
import json
import logging
import os
import pytz


logger = getLogger('ftw.contentstats')
root_logger = logging.root

LOG_TZ = get_localzone()


def setup_logger():
    """Set up logger that writes to a JSON logfile or a Fluentd instance.

    May be invoked multiple times, and must therefore be idempotent.
    """
    if not logger.handlers:
        fluent_host = os.environ.get('FLUENT_HOST')
        fluent_port = int(os.environ.get('FLUENT_PORT', 24224))

        if fluent_host:
            setup_fluent_handler(fluent_host, fluent_port)
        else:
            setup_jsonfile_handler()

        logger.setLevel(logging.INFO)
        logger.propagate = False

    return logger


def setup_fluent_handler(fluent_host, fluent_port):
    """Set up handler that writes to a Fluentd / Fluent Bit instance.
    """
    tag = 'contentstats-json.log'
    ns = os.environ.get('KUBERNETES_NAMESPACE')
    if ns:
        tag = '-'.join((ns, tag))

    handler = FluentHandler(tag, fluent_host, fluent_port)
    handler.setFormatter(FluentRecordFormatter())
    logger.addHandler(handler)
    return logger


def setup_jsonfile_handler():
    """Set up handler that writes to the JSON based logfile.
    """
    path = get_logfile_path()
    if path is not None:
        handler = FileHandler(path)
        logger.addHandler(handler)
    return logger


def get_log_dir_path():
    """Determine the path of the deployment's var/log/ directory.

    This will be derived from Zope2's EventLog location, in order to not
    have to figure out the path to var/log/ ourselves from buildout.
    """
    zconf = getConfiguration()
    eventlog = getattr(zconf, 'eventlog', None)
    if eventlog is None:
        root_logger.error('')
        root_logger.error(
            "ftw.contentstats: Couldn't find eventlog configuration in order "
            "to determine logfile location!")
        root_logger.error(
            "ftw.contentstats: No content stats logfile will be written!")
        root_logger.error('')
        return None

    handler_factories = eventlog.handler_factories
    eventlog_path = handler_factories[0].section.path
    log_dir = dirname(eventlog_path)
    return log_dir


def get_logfile_path():
    """Determine the path for our JSON log.
    """
    log_dir = get_log_dir_path()
    path = join(log_dir, 'contentstats-json.log')
    return path


def log_stats_to_file():
    logger = setup_logger()

    stats = ContentStats().get_raw_stats()

    ts = datetime.utcnow().replace(tzinfo=pytz.utc)
    stats['timestamp'] = ts.astimezone(LOG_TZ).isoformat()
    stats['site'] = get_site_id()

    logger.info(json.dumps(stats, sort_keys=True))


def get_site_id():
    site = getSite()
    if site:
        return site.id
    return ''



from ftw.contentstats.logger import log_stats_to_file
from zExceptions import Unauthorized
from zope.publisher.browser import BrowserView


class DumpContentStatsView(BrowserView):
    """Dumps content statistics to a logfile.

    This view is accessible to Anonymous (zope.Public), but *only* when
    requested from localhost!

    This is so it can easily be triggered from cron jobs or scripts running
    on the same host as the Plone site, without the need to store credentials.
    """

    def _verify_access(self):
        # Only allow access from localhost
        src_ip = self.request.getClientAddr()
        if src_ip != '127.0.0.1':
            raise Unauthorized()

    def __call__(self):
        self.request.response.setHeader('Content-Type', 'text/plain')
        self._verify_access()

        if self.request.method.upper() != 'POST':
            self.request.response.setStatus(405)
            return 'Method not allowed, please POST'

        log_stats_to_file()
        self.request.response.setStatus(204)
        return ''



from ftw.contentstats.interfaces import IStatsProvider
from ftw.contentstats.logger import get_log_dir_path
from ftw.contentstats.tests import assets
from ftw.contentstats.tests import FunctionalTestCase
from os.path import join
from zope.component import getMultiAdapter
import os


class TestDiskUsageProvider(FunctionalTestCase):

    def setUp(self):
        super(TestDiskUsageProvider, self).setUp()
        self.grant('Manager')

        self.provider = getMultiAdapter((self.portal, self.portal.REQUEST),
                                        IStatsProvider,
                                        name='disk_usage')

        log_dir = get_log_dir_path()
        self.disk_usage_path = join(log_dir, 'disk-usage.json')

    def tearDown(self):
        super(TestDiskUsageProvider, self).tearDown()
        try:
            os.unlink(self.disk_usage_path)
        except OSError:
            pass

    def test_disk_usage_empty_if_file_not_present(self):
        disk_usage = self.provider.get_raw_stats()
        self.assertEqual({}, disk_usage)

    def test_disk_usage_getting_read_from_json_file(self):
        with open(self.disk_usage_path, 'w') as disk_usage_file:
            disk_usage_file.write(assets.load('disk-usage.json'))

        disk_usage = self.provider.get_raw_stats()
        self.assertEqual({'blobstorage': 45,
                          'filestorage': 20,
                          'total': 1024},
                         disk_usage)



from zope.interface import Interface


class IStatsProvider(Interface):
    """Interface for stats provider adapters
    """

    def __init__(context, request):
        """Adapts context and request, context is usually a Plone site"""

    def title():
        """Return a human readable title of the stats provider
        """

    def get_raw_stats():
        """Extract and return raw stats.

        Return value is a dict with key:value pairs, where key should be a
        stable, internal ID. If a different display name is desired, a mapping
        should be provided by implementing get_display_names().
        """

    def get_display_names():
        """Return a key: display_name mapping of human readable key names.

        If no alternate display names are needed, this should return None.
        """


class IStatsKeyFilter(Interface):
    """Interface for named adapters that allow stats providers to filter their
    returned stats data by key.

    The adapter name indicates what type of keys it's supposed to be applied
    to by stats providers, e.g. 'portal_types', 'review_states', ...
    """

    def __init__(context, request):
        """Adapts context and request, context is usually a Plone site"""

    def keep(key):
        """Returns ``True`` if the key in question should be kept, ``False``
        otherwise.
        """



from ftw.builder import Builder
from ftw.builder import create
from ftw.contentstats.interfaces import IStatsProvider
from ftw.contentstats.tests import FunctionalTestCase
from zope.component import getMultiAdapter


class TestReviewStatesProvider(FunctionalTestCase):

    def setUp(self):
        super(TestReviewStatesProvider, self).setUp()
        self.grant('Manager')

        self.provider = getMultiAdapter((self.portal, self.portal.REQUEST),
                                        IStatsProvider,
                                        name='review_states')

    def create_content(self):
        self.set_workflow_chain('Document', 'simple_publication_workflow')
        self.set_workflow_chain('Folder', 'simple_publication_workflow')
        create(Builder('folder'))
        create(Builder('page'))
        create(Builder('page')
               .in_state('published'))

    def test_review_states_counts_empty(self):
        counts = self.provider.get_raw_stats()
        self.assertEqual({}, counts)

    def test_review_states_counts_reported_correctly(self):
        self.create_content()
        counts = self.provider.get_raw_stats()
        self.assertEqual({'private': 2, 'published': 1}, counts)

    def test_display_names_reported_correctly(self):
        self.create_content()
        titles = self.provider.get_display_names()
        self.assertDictContainsSubset({
            'private': u'private',
            'published': u'published'},
            titles)

    def test_respects_review_states_filter(self):
        self.create_content()
        self.load_zcml_string('<include package="ftw.contentstats.demo" />')
        counts = self.provider.get_raw_stats()
        self.assertEqual({'published': 1}, counts)
